/*************************************************************************\
 * File   : TkImageTools.cxx
 * Purpose: Resize and other image related tools
 *
 * Author : Tom Wilkason
 * Dated  : 12/22/2005
 *
 * Revision History:
 * -----------------
 * $Revision:$
 * $Log:$
\************************************************************************/

#include <stdio.h>
#include <tcl.h>
#include <tk.h>
static int imageResize  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
void filtering3x3 (Tk_PhotoImageBlock* db, Tk_PhotoImageBlock* sb, int *m);
int createTkCommands(Tcl_Interp *interp)
{
   Tcl_CreateObjCommand(interp, "tkImageTools::resize",   imageResize,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   return 1;
}
/* ---------------------------------------------------------------------
 *
 *  ScalePhotoBlock --
 *
 *      Scales a photo block. blockPtr is modified to hold new photo
 *      block information.
 *      On success, once the block is used (passed to Tk_PhotoPutBlock()),
 *      blockPtr->pixelPtr must be freed.
 *      Ported from PBM by Eric Boudaillier http://wiki.tcl.tk/10504
 *      Tom Wilkason wrapped it into a TCL command
 * ---------------------------------------------------------------------
 */
int
  ScalePhotoBlock (
                  Tcl_Interp *interp,
                  Tk_PhotoImageBlock *srcBlockPtr,
                  Tk_PhotoImageBlock *dstBlockPtr,
                  int newcols,
                  int newrows)
{
#define SCALE      4096
#define HALFSCALE  2048
#define MAXPIXVAL  255
   unsigned char *xelrow;
   unsigned char *tempxelrow;
   unsigned char *newimage;
   unsigned char *newxelrow;
   register unsigned char *xP;
   register unsigned char *nxP;
   int rows, cols, srcrow, bpp;
   register int row, col, needtoreadrow, p;
   double xscale, yscale;
   long   sxscale, syscale;
   register long fracrowtofill, fracrowleft;
   long  *pixarray[4], v;
   int pitch, newpitch;

   cols    = srcBlockPtr->width;
   rows    = srcBlockPtr->height;
   bpp     = srcBlockPtr->pixelSize;
   xscale = (double) newcols / cols;
   yscale = (double) newrows / rows;
   sxscale = (long) (xscale * SCALE);
   syscale = (long) (yscale * SCALE);

   pitch = srcBlockPtr->pitch;
   newpitch = newcols*bpp;

   if (newrows != rows)
   {
      tempxelrow = (unsigned char*) Tcl_AttemptAlloc(bpp*cols);
      if (!tempxelrow)
      {
         Tcl_SetResult(interp, "memory allocation failed", TCL_STATIC);
         return TCL_ERROR;
      }
   } else
   {
      tempxelrow = 0;
   }
   pixarray[0] = (long*) Tcl_AttemptAlloc(bpp*cols*sizeof(long));
   if (!pixarray[0])
   {
      if (tempxelrow)
         Tcl_Free((char*) tempxelrow);
      Tcl_SetResult(interp, "memory allocation failed", TCL_STATIC);
      return TCL_ERROR;
   }
   pixarray[1] = pixarray[0] + cols;
   pixarray[2] = pixarray[1] + cols;
   pixarray[3] = pixarray[2] + cols;
   for (p = 0; p < bpp; ++p)
   {
      for (col = 0; col < cols; ++col)
         pixarray[p][col] = HALFSCALE;
   }

   fracrowleft   = syscale;
   needtoreadrow = 1;
   fracrowtofill = SCALE;

   newimage = (unsigned char*) Tcl_AttemptAlloc(newcols * newrows * bpp);
   if (!newimage)
   {
      if (newrows != rows)
         Tcl_Free((char*) tempxelrow);
      Tcl_Free((char*) pixarray[0]);
      Tcl_SetResult(interp, "couldn't allocate image", TCL_STATIC);
      return TCL_ERROR;
   }
   newxelrow = newimage;
   xelrow = 0;
   for (row = 0, srcrow = 0; row < newrows; ++row)
   {
        /* First scale Y from xelrow into tempxelrow. */
      if (newrows == rows)  /* shortcut Y scaling if possible */
      {
         tempxelrow = xelrow = srcBlockPtr->pixelPtr + (pitch * srcrow++);
      } else
      {
         while (fracrowleft < fracrowtofill)
         {
            if (needtoreadrow && srcrow < rows)
               xelrow = srcBlockPtr->pixelPtr + (pitch * srcrow++);
            for (col = 0, xP = xelrow; col < cols; ++col)
               for (p = 0; p < bpp; ++p, ++xP)
                  pixarray[p][col] += fracrowleft * *xP;
            fracrowtofill -= fracrowleft;
            fracrowleft = syscale;
            needtoreadrow = 1;
         }

            /* Now fracrowleft is >= fracrowtofill, so we can produce a row. */
         if (needtoreadrow && srcrow < rows)
         {
            xelrow = srcBlockPtr->pixelPtr + (pitch * srcrow++);
            needtoreadrow = 0;
         }
         for (col = 0, xP = xelrow, nxP = tempxelrow; col < cols; ++col)
         {
            for (p = 0; p < bpp; ++p, ++xP, ++nxP)
            {
               v = (pixarray[p][col] + fracrowtofill * *xP) / SCALE;
               *nxP = (unsigned char) (v > MAXPIXVAL ? MAXPIXVAL:v);
               pixarray[p][col] = HALFSCALE;
            }
         }
         fracrowleft -= fracrowtofill;
         if (fracrowleft == 0)
         {
            fracrowleft = syscale;
            needtoreadrow = 1;
         }
         fracrowtofill = SCALE;
      }


        /* Now scale X from tempxelrow into newxelrow and write it out. */
      if (newcols == cols)  /* shortcut X scaling if possible */
      {
         for (col = 0, xP = tempxelrow, nxP = newxelrow; col < cols; ++col)
            for (p = 0; p < bpp; ++p, ++xP, ++nxP)
               *nxP = *xP;
         newxelrow = newxelrow + newpitch;
      } else
      {
         long pixval[4];
         register long fraccoltofill, fraccolleft;
         register int needcol;

         nxP = newxelrow;
         fraccoltofill = SCALE;
         for (p = 0; p < bpp; ++p)
            pixval[p] = HALFSCALE;
         needcol = 0;
         for (col = 0, xP = tempxelrow; col < cols; ++col, xP += bpp)
         {
            fraccolleft = sxscale;
            while (fraccolleft >= fraccoltofill)
            {
               if (needcol)
               {
                  nxP += bpp;
                  for (p = 0; p < bpp; ++p)
                     pixval[p] = HALFSCALE;
               }
               for (p = 0; p < bpp; ++p)
               {
                  pixval[p] = (pixval[p] + fraccoltofill * xP[p]) / SCALE;
                  if (pixval[p] > MAXPIXVAL) pixval[p] = MAXPIXVAL;
                  nxP[p]    = (unsigned char) pixval[p];
               }
               fraccolleft  -= fraccoltofill;
               fraccoltofill = SCALE;
               needcol = 1;
            }
            if (fraccolleft > 0)
            {
               if (needcol)
               {
                  nxP += bpp;
                  for (p = 0; p < bpp; ++p)
                     pixval[p] = HALFSCALE;
                  needcol = 0;
               }
               for (p = 0; p < bpp; ++p)
                  pixval[p] += fraccolleft * xP[p];
               fraccoltofill -= fraccolleft;
            }
         }
         if (fraccoltofill > 0)
         {
            xP -= bpp;
            for (p = 0; p < bpp; ++p)
               pixval[p] += fraccoltofill * xP[p];
         }
         if (!needcol)
         {
            for (p = 0; p < bpp; ++p)
            {
               pixval[p] /= SCALE;
               if (pixval[p] > MAXPIXVAL) pixval[p] = MAXPIXVAL;
               nxP[p] = (unsigned char) pixval[p];
            }
         }
         newxelrow = newxelrow + newpitch;
      }
   }

   Tcl_Free((char*) pixarray[0]);
   if (newrows != rows)
      Tcl_Free((char*) tempxelrow);

   dstBlockPtr->width = newcols;
   dstBlockPtr->height = newrows;
   dstBlockPtr->pixelSize = bpp;
   dstBlockPtr->pitch = newpitch;
   dstBlockPtr->pixelPtr = newimage;
   dstBlockPtr->offset[0] = 0;
   dstBlockPtr->offset[1] = 1;
   dstBlockPtr->offset[2] = 2;
   dstBlockPtr->offset[3] = 3;

   return TCL_OK;

#undef SCALE
#undef HALFSCALE
#undef MAXPIXVAL
}


/* Unused for now */
static int getp(Tk_PhotoImageBlock* b, int x, int y, int* rgb)
{
   unsigned char* pixelp;

   if (x < 0 || x >= b->width || y < 0 || y >= b->height)
   {
      return 0;
   }
   pixelp = (b->pixelPtr) + y * (b->pitch) + x * (b->pixelSize);
   * (rgb+0) = * (pixelp+(b->offset[0]));
   * (rgb+1) = * (pixelp+(b->offset[1]));
   * (rgb+2) = * (pixelp+(b->offset[2]));

   return 1;
}

/* Unused for now */
static int setp(Tk_PhotoImageBlock* b, int x, int y, int* rgb)
{
   unsigned char* pixelp;

   if (x < 0 || x >= b->width || y < 0 || y >= b->height)
   {
      return 0;
   }
   pixelp = b->pixelPtr + y * (b->pitch) + x * (b->pixelSize);
   * (pixelp+(b->offset[0])) = (unsigned char)(* (rgb + 0));
   * (pixelp+(b->offset[1])) = (unsigned char)(* (rgb + 1));
   * (pixelp+(b->offset[2])) = (unsigned char)(* (rgb + 2));

   return 1;
}

/*----------------------------------------------------------------------*/
/* Unused for now */
static void filtering3x3(
                        Tk_PhotoImageBlock* db, Tk_PhotoImageBlock* sb, int m[3][3])
{
   int x, y, sx, sy;
   int rgb[3];

    /* ϽǤʤΤǥˤ */
   rgb[0]=rgb[1]=rgb[2]=0;
   for (y=0; y<(sb->height); y++)
   {
      setp(db, 0, y, rgb);
      setp(db, sb->width-1, y, rgb);
   }
   for (x=1; x<(sb->width-1); x++)
   {
      setp(db, x, 0, rgb);
      setp(db, x, db->height-1, rgb);
   }
    /* ե륿 */
   for (y=1; y<(sb->height-1); y++)
   {
      for (x=1; x<(sb->width-1); x++)
      {
         int r, g, b, ctr;
         r=0; g=0; b=0; ctr=0;
         for (sy=0; sy<3; sy++)
         {
            for (sx=0; sx<3; sx++)
            {
               if (getp(sb, x+sx-1, y+sy-1, rgb) == 1)
               {
                  r += m[sy][sx] * rgb[0];
                  g += m[sy][sx] * rgb[1];
                  b += m[sy][sx] * rgb[2];
                  ctr+=(m[sy][sx]>=0)? m[sy][sx]:-m[sy][sx];
               }
            }
         }
         if (r<0) r=-r;
         if (g<0) g=-g;
         if (b<0) b=-b;
         rgb[0] = (int)(r / ctr);
         rgb[1] = (int)(g / ctr);
         rgb[2] = (int)(b / ctr);
         setp(db, x, y, rgb);
      }
   }
}

/*----------------------------------------------------------------------*/

/* Unused for now */
void copy(Tk_PhotoImageBlock *db, Tk_PhotoImageBlock *sb)
{
   int x, y;
   int rgb[3];

   for (y=0; y<sb->height; y++)
   {
      for (x=0; x<sb->width; x++)
      {
         if (getp(sb, x, y, rgb) == 1) setp(db, x, y, rgb);
      }
   }
}
/*----------------------------------------------------------------------*
 * Tcl/Tk imageResize command
 *
 *   Usage: image src dest xscale yscale
 *   Tom Wilkason
 *----------------------------------------------------------------------*/
static int
  imageResize
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{

   if (objc < 5)
   {
      Tcl_WrongNumArgs(interp, 1, objv, "sourceImg destImg newRows newCols");
      return TCL_ERROR;
   }
   int result;
   char *src, *dest;
   Tk_PhotoHandle sp, dp;
   Tk_PhotoImageBlock sb, db;
   int newrows=1;
   int newcols=1;
   src = Tcl_GetString(objv[1]);
   dest = Tcl_GetString(objv[2]);
   Tcl_GetIntFromObj(interp,objv[3],&newrows);
   Tcl_GetIntFromObj(interp,objv[4],&newcols);
   /* Get images */
   sp = Tk_FindPhoto(interp, src );
   dp = Tk_FindPhoto(interp, dest);
   if (sp==0)
   {
      Tcl_AppendResult (interp, "Invalid image source specified",NULL);
      return TCL_ERROR;
   } else if (dp==0)
   {
      Tcl_AppendResult (interp, "Invalid image destination specified",NULL);
      return TCL_ERROR;
   }
   db.pixelPtr = 0;
   Tk_PhotoGetImage(sp, &sb);
   Tk_PhotoGetImage(dp, &db);
   result = ScalePhotoBlock(interp,&sb,&db,newrows,newcols);
   if (result == TCL_OK)
   {
      /* void under 8.3 not 8.4? */
      /*
         The compRule parameter to Tk_PhotoPutBlock specifies a compositing rule
         that says what to do with transparent pixels.  The value
         TK_PHOTO_COMPOSITE_OVERLAY says that the previous contents of the photo
         image should show through, and the value TK_PHOTO_COMPOSITE_SET says that
         the previous contents of the photo image should be completely ignored, and
         the values from the block be copied directly across.  The behavior in
         Tk8.3 and earlier was equivalent to having TK_PHOTO_COMPOSITE_OVERLAY as a
         compositing rule.
      */
      Tk_PhotoPutBlock(dp, &db,0, 0, newrows, newcols,TK_PHOTO_COMPOSITE_OVERLAY);
      if (db.pixelPtr)
      {
         Tcl_Free((char *)db.pixelPtr);
      }
      /*
      if (result != TCL_OK)
      {
         Tcl_AppendResult (interp, "Error putting image data into destination",NULL);
         return TCL_ERROR;
      }
      */
      return TCL_OK;
   } else {
      return result;
   }
}

